# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import datetime
import logging
import multiprocessing
from dataclasses import dataclass
from multiprocessing.sharedctypes import Synchronized
from typing import List, Optional, Tuple, Union

from logger.comm import Comm, CommandList
from logger.measure.data import MeasureData


@dataclass
class ReaderArgs:
    """読み込みプロセスに渡す設定情報の抽象クラス

    Sensorを継承したクラスでは、このクラスを継承したクラスを実装しto_reader_argsで返す
    """

    model: str
    serial: str
    port: str
    baud: int
    start_command: CommandList
    end_command: CommandList
    record_size: int
    record_begin: int
    record_end: int
    record_per_sec: int
    count_diff: int
    count_max: int
    count_start: int
    read_length: int = Comm.DEFAULT_READ_BYTES
    timeout: float = Comm.DEFAULT_TIMEOUT

    def format_packet(self, index: int, packet: List[int]) -> MeasureData:
        """1レコード分のデータを受け取って計測データに変換するメソッド

        ReaderArgsを継承するクラスで具体的な処理を記載する

        Args:
            index (int): 何番目のデータか
            packet (List[int]): 変換対象のデータ配列

        Raises:
            NotImplementedError: サブクラスで実装しなかった場合

        Returns:
            MeasureData: 変換後のpythonデータクラス
        """
        raise NotImplementedError

    def complement_data(
        self, prev: Optional[MeasureData], next: MeasureData
    ) -> List[MeasureData]:
        """カウント値をチェックしてデータを補完するメソッド

        Args:
            prev (Optional[MeasureData]): 補完済みの最後のデータ
            next (MeasureData): 補完前の新たに配列に追加するデータ

        Returns:
            List[MeasureData]: データ欠損を補完した計測データ配列
        """
        logger = logging.getLogger(
            f"{__name__}.{self.model}.{self.serial}.complement_data"
        )

        # 前のデータがない時は1番最初のデータを検証する時なので-1番目のデータをprevとしてセットする
        if prev is None:
            prev = MeasureData(
                index=-1,
                count=self.count_start - self.count_diff,
                temperature=0,
                x=0,
                y=0,
                z=0,
                flag=0,
            )

        # 次のカウント値に一致するまで一回ごとの差分を足し続けて差分を計算する
        diff = self.count_diff
        while (diff + prev.count) % self.count_max != next.count:
            diff += self.count_diff

        # カウントの差分と想定される差分が異ならない場合はデータ損失がないためそのまま返す
        if diff == self.count_diff:
            next.index = prev.index + 1
            return [next]

        result: List[MeasureData] = []
        # 欠損データを補完する
        logger.warning(
            f"Missing {(diff // self.count_diff) - 1} data from index: {prev.index + 1}. Complement them."
        )
        for _ in range(0, diff, self.count_diff):
            index = prev.index + 1
            count = (prev.count + self.count_diff) % self.count_max

            # カウント値がnextのものと一致しているときはループの終端に来ているので元のデータを入れる
            if count == next.count:
                next.index = index
                result.append(next)
            else:
                complement = self._gen_invalid_data(index=index, count=count)
                result.append(complement)
                prev = complement

        return result

    def _gen_invalid_data(self, index: int, count: int) -> MeasureData:
        """データ抜けがあった際に不正な値であることを示すデータクラスオブジェクトを生成するメソッド

        Args:
            index (int): 不正な計測データのindex
            count (int): 不正な計測データのカウント値、complement_dataの中で前のデータから算出する

        Raises:
            NotImplementedError: サブクラスで実装しなかった場合

        Returns:
            MeasureData: 補完したindexとcountを持つ不正なことを表すデータクラスオブジェクト
        """
        raise NotImplementedError


@dataclass
class WriterArgs:
    """読み込みプロセスに渡す設定情報のクラス

    Sensorを継承したクラスでは、to_writer_argsでこのクラスオブジェクトを返す
    """

    model: str
    serial: str
    logger_id: str
    port: str
    record_per_file: int

    def get_file_name(self, start_time: datetime.datetime) -> str:
        """与えられた時刻でデータを出力するcsvのファイル名を生成するメソッド

        Args:
            start_time (datetime.datetime): 出力開始時刻

        Returns:
            str: ディレクトリ名を含まないcsvファイルのファイル名
        """
        return f"{self.model}_{self.serial}_{self.logger_id}_{self.port[-1]}_{start_time:%y%m%d_%H%M%S}.csv"


def _packets_to_data(
    packets: List[int], args: ReaderArgs
) -> Tuple[List[MeasureData], List[int]]:
    """センサーから読み込んだデータから計測データクラスオブジェクトの配列に変換する関数

    Args:
        packets (List[int]): センサーから読み込んだデータ
        args (ReaderArgs): 計測設定情報を保持するオブジェクト

    Returns:
        Tuple[List[MeasureData], List[int]]: 変換が終わった計測データの配列と1レコード分に足りず変換ができなかった残ったデータ
    """
    logger = logging.getLogger(f"{__name__}.{args.model}.{args.serial}.packets_to_data")
    index = 0
    data: List[MeasureData] = []
    rest: List[int] = []
    while True:
        # データずれが発生した時にWarningログを出力どうかのフラグ
        should_log = True
        packet = packets[index : index + args.record_size]

        # packetサイズが0になっている恐れもあるため、while文の条件として長さを判定する
        while len(packet) == args.record_size and (
            packet[0] != args.record_begin or packet[-1] != args.record_end
        ):
            if should_log:
                logger.warning(
                    f"Invalid packet boundary: beg={packet[0]:#X}, end={packet[-1]:#X}. Fix packet."
                )
                # 一回のパケットずれごとに一度出力したらもう出力しないようにする
                should_log = False
            # indexを1つずらして再調査
            index += 1
            packet = packets[index : index + args.record_size]

        # 切り取ったpacketが1レコード分に満たない時はデータの終端に来ているのでループを抜ける
        if len(packet) < args.record_size:
            rest = packet
            break

        data.append(args.format_packet(0, packet))

        # indexをレコード分だけずらして次のループへ
        index += args.record_size

    return data, rest


def reader_job(
    queue: multiprocessing.Queue,
    error_queue: multiprocessing.Queue,
    args: ReaderArgs,
    measuring: Synchronized,
) -> None:
    """サブプロセス上でデバイスからのデータを受け取り、整形して読み込みプロセスに渡す関数

    Args:
        queue (multiprocessing.Queue): 計測データを読み込みプロセスに渡すためのQueue
        measuring (Synchronized): 計測中かどうかのフラグ
        error_queue (multiprocessing.Queue): エラーが発生した際に親プロセスに渡すためのQueue
        args (ReaderArgs): 計測設定情報を保持するオブジェクト
    """
    logger = logging.getLogger(f"{__name__}.{args.model}.{args.serial}.reader")
    comm = Comm()
    try:
        comm.open(args.port, args.baud, args.timeout)

        # 計測開始コマンド送信
        comm.send(args.start_command)
        send: List[MeasureData] = []
        data: List[int] = []
        process_count = 0
        # 補完ずみの最後のデータ
        last_data: Optional[MeasureData] = None

        while measuring.value:
            result = comm.read(args.read_length)
            data.extend(result)

            # パケットから計測データに変換する
            converted, rest = _packets_to_data(data, args)
            data = rest

            # 計測データを補完しながら送信するためのデータに詰めていく
            for d in converted:
                send.extend(args.complement_data(last_data, d))
                last_data = send[-1]

            # 1秒分以上のデータが溜まっていたら書き込みプロセスに送りつけるため送出
            if len(send) >= args.record_per_sec:
                queue.put(send[0 : args.record_per_sec])
                send = send[args.record_per_sec :]
                process_count += args.record_per_sec

        # 計測中フラグがオフになったら残りを送信する
        queue.put(send)

    except Exception as e:
        logger.exception("An error occurred during reading data")
        error_queue.put(JobError(model=args.model, serial=args.serial, error=e))
    finally:
        try:
            # 終了コマンドを送信
            comm.send(args.end_command)
        except Exception:
            logger.exception("Failed to send stop command")
        try:
            comm.close()
        except Exception:
            logger.exception("Failed to disconnect device connection")
        try:
            # writer_processに終了を通知
            queue.put(None)
        except Exception:
            # reader_processが終了しなくなるのはかなりまずいのでfatalで出す
            logger.fatal("Failed to notify finishing writer process")


def writer_job(
    queue: multiprocessing.Queue,
    error_queue: multiprocessing.Queue,
    args: WriterArgs,
    output_dir: str,
) -> None:
    """サブプロセス上で読み取りプロセスから受け取ったデータをファイルに書き込む関数

    Args:
        queue (multiprocessing.Queue): 読み取りプロセスからデータを受け取るためのQueue
        error_queue (multiprocessing.Queue): エラーが発生した際に親プロセスに渡すためのQueue
        output_dir (str): 計測情報の保存ディレクトリ
        args (WriterArgs): 計測設定情報を保持するオブジェクト
    """
    logger = logging.getLogger(f"{__name__}.{args.model}.{args.serial}.writer")
    # 初回書き込み設定
    output_file = f"{output_dir}/{args.get_file_name(datetime.datetime.now())}"
    mode = "w"
    write_count = 0
    # try句がエラーを起こさずに正常に完了したか
    success = False
    logger.info(f"Initial output to {output_file}")
    try:
        while True:
            # 書き込み側の処理間隔を広げるため、1秒分のレコードを読み取り側でまとめて送出している
            records: Union[List[MeasureData], None] = queue.get()
            # readerからNoneを受け取ったら終了の印
            if records is None:
                success = True
                break

            # 計測終了時の送出はレコード数が0の可能性があるため空文字列を書き込まないようにする
            if len(records) > 0:
                lines = "".join((d.to_csv_line() for d in records))
                with open(
                    output_file, mode=mode, encoding="UTF-8", newline="\r\n"
                ) as f:
                    f.write(lines)
                    logger.debug(f"write {len(records)} lines")
                write_count += len(records)
                # 書き込むたびにファイル書き込みのモードをappendにする
                mode = "a"

            # 1ファイルあたりのレコード数を書き込み終えたら書き込み設定を初期化する
            if write_count >= args.record_per_file:
                output_file = (
                    f"{output_dir}/{args.get_file_name(datetime.datetime.now())}"
                )
                mode = "w"
                write_count = 0
                logger.info(f"Output file is rotated to {output_file}")
    except Exception as e:
        logger.exception("An error occurred during writing data")
        # 例外発生時はqueueを通じて親プロセスに例外を送出する
        error_queue.put(JobError(model=args.model, serial=args.serial, error=e))
    finally:
        # reader processからの終了通知を受け取る前にtry句を抜けていたらreader processからの終了通知を受け取るまでqueueから取得続ける
        if not success:
            try:
                while True:
                    # この条件節に入った場合は書き込みに失敗している可能性が高いため、None以外のレコードは無視して良い
                    message: Union[List[MeasureData], None] = queue.get()
                    if message is None:
                        break
            except Exception:
                logger.fatal("Failed to get reader process finish message")


class JobError:
    """サブプロセス上で発生したエラーをどのセンサーで発生したか親プロセスで識別できるように情報を付与したクラス"""

    def __init__(self, model: str, serial: str, error: Exception) -> None:
        self.model = model
        self.serial = serial
        self.error = error
