# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import logging
import multiprocessing
from multiprocessing.sharedctypes import Synchronized
from typing import Optional

from logger.comm import CommandList, WaitCommand, send
from logger.measure.job import ReaderArgs, WriterArgs, reader_job, writer_job
from logger.process import ApplicationProcess
from logger.utils.convert import to_uint16


class Sensor:
    """センサー抽象クラス"""

    DEFAULT_START_COMMAND: CommandList = [
        [0, 0xFF, 0xFF, 0x0D],  # 復帰おまじない3回
        [0, 0xFF, 0xFF, 0x0D],
        [0, 0xFF, 0xFF, 0x0D],
        [0, 0xFE, 0x00, 0x0D],  # WINDOW_ID(L) write command.(WINDOW=0)
        [0, 0x83, 0x01, 0x0D],  # 自動サンプリングモード
    ]
    DEFAULT_END_COMMAND: CommandList = [
        [0, 0xFE, 0x00, 0x0D],  # Window 0
        [0, 0x83, 0x02, 0x0D],  # MODE_CTRL: 1-0:10=Config移行
        WaitCommand(0.01),  # Wait > 1ms
    ]
    DEFAULT_RECORD_BEGIN = 0x80
    DEFAULT_RECORD_END = 0x0D

    def __init__(
        self,
        model: str,
        port: str,
        baud: int,
        product_id: str,
        sps: int,
        logger_id: str,
        file_rotate_min: int,
        mode: str,
        physical: str,
        filter: Optional[str] = None,
        rms_pp_interval: Optional[int] = None,
    ) -> None:
        self._logger = logging.getLogger(__name__)
        self.model = model
        self.port = port
        self.baud = baud
        self.product_id = product_id
        self.serial = ""
        self.firm_ver = 0
        self.sps = sps
        self.logger_id = logger_id
        self.file_rotate_min = file_rotate_min
        self.mode = mode
        self.physical = physical
        self.filter = filter
        self.rms_pp_interval = rms_pp_interval

        self._reader_process: Optional[multiprocessing.Process] = None
        self._writer_process: Optional[multiprocessing.Process] = None
        self._measuring: Synchronized = multiprocessing.Value("b", False)  # type: ignore
        self._queue: Optional[multiprocessing.Queue] = None

    def __str__(self) -> str:
        return f"{self.model} #{self.serial} @{self.port}"

    def ready(self) -> bool:
        """センサーに異常がなく計測を開始できるか確認するメソッド

        共通であることを確認: A352

        Returns:
            bool: センサーの異常有無
        """
        commands: CommandList = [
            # W: SOFT_RST
            [0, 0xFE, 0x01, 0x0D],  # Window 1
            [0, 0x8A, 0x80, 0x0D],  # GLOB_CMD: 7:SOFT_RST=1
            WaitCommand(1.2),
            # R: GLOB_CMD
            [0, 0xFE, 0x01, 0x0D],  # Window 1
            [4, 0x0A, 0x00, 0x0D],  # R: GLOB_CMD -> 0-3byte
            # R: DIAG_STAT
            [0, 0xFE, 0x00, 0x0D],  # Window 0
            [4, 0x04, 0x00, 0x0D],  # R: DIAG_STAT -> 4-7byte
        ]

        # 送信
        result = send(commands, self.port, self.baud)

        # 結果解析
        success = True

        # GLOB_CMD.NOT_READY[10] == 0
        #   ↑返却16bit中、10bit値を検査
        val = result[0:4][1]
        if (val & 0b00000100) != 0:
            success = False
            self._logger.error(f"Sensor device: NOT_READY ({bin(val)})")

        # DIAG.HARD_ERR[7-5] == 000
        #   ↑返却16bit中、7-5bit値を検査
        val = result[4:8][2]
        if (val & 0b11100000) != 0:
            success = False
            self._logger.error(f"Sensor device: HARD_ERR ({bin(val)})")

        return success

    def load_serial(self) -> None:
        """シリアルナンバーを読み込み、保持するメソッド

        共通であることを確認: A352
        """
        # シリアルコマンドリスト
        commands: CommandList = [
            [0, 0xFE, 0x01, 0x0D],  # Window 1
            [4, 0x74, 0x01, 0x0D],  # Serial_Number1
            [4, 0x76, 0x01, 0x0D],  # Serial_Number2
            [4, 0x78, 0x01, 0x0D],  # Serial_Number3
            [4, 0x7A, 0x01, 0x0D],  # Serial_Number4
        ]

        # コマンド送信
        result = send(commands, self.port, self.baud)

        # シリアル文字列合成
        #   返却は [Addr, Serial2, Serial1, CR] の構成
        serial = ""
        for i in range(0, len(result), 4):
            serial += chr(result[i + 2])  # Serial1 を追加
            serial += chr(result[i + 1])  # Serial2 を追加

        # シリアルの保持
        self._logger.debug(f"load serial: {serial}")
        self.serial = serial

    def load_firm_ver(self) -> None:
        """ファームウェアバージョンを読み込み、保持するメソッド

        共通であることを確認: A352
        """
        # シリアルコマンドリスト
        commands: CommandList = [
            [0, 0xFE, 0x01, 0x0D],  # Window 1
            [4, 0x72, 0x00, 0x0D],  # FIRMVER
        ]

        # コマンド送信
        result = send(commands, self.port, self.baud)
        # TODO: 仕様書だと16bit整数値になっているがData Loggerの出力値が異なる。大森さんに確認する。
        firm_ver = to_uint16(result[1], result[2])

        # ファームウェアバージョンの保持
        self._logger.debug(f"load firm_ver: {firm_ver}")
        self.firm_ver = firm_ver

    def init(self) -> None:
        """与えられた計測設定に基づきセンサーの計測情報を設定するメソッド

        サブクラスは具体的な処理を実装する

        Raises:
            NotImplementedError: サブクラスで実装されていない場合
        """
        raise NotImplementedError

    def start(self, error_queue: multiprocessing.Queue, output_dir: str) -> None:
        """計測を開始するメソッド

        サブクラスはto_reader_argsと_to_writer_argsを実装すること

        Args:
            error_queue (multiprocessing.Queue): サブプロセス内で発生したエラーを伝えるQueue
            output_dir (str): 計測情報の保存ディレクトリ
        """
        self._measuring.value = True
        # queueは計測間で使いまわさないため初期化する
        self._queue = multiprocessing.Queue(-1)
        self._reader_process = ApplicationProcess(
            target=reader_job,
            args=(self._queue, error_queue, self.to_reader_args(), self._measuring),
        )
        self._writer_process = ApplicationProcess(
            target=writer_job,
            args=(self._queue, error_queue, self.to_writer_args(), output_dir),
        )
        self._reader_process.start()
        self._writer_process.start()

    def stop(self) -> None:
        """計測を終了するメソッド"""
        self._measuring.value = False
        # queueから全てのデータを消費する必要があるため、writer_processを先に終了待ちする
        if self._writer_process:
            self._writer_process.join()
            self._writer_process = None
        if self._reader_process:
            self._reader_process.join()
            self._reader_process = None
        # 計測で使い終わったqueueをcloseする。
        if self._queue:
            self._queue.close()
            self._queue = None

    def to_reader_args(self) -> ReaderArgs:
        """読み込みプロセスに渡す設定情報に変換するメソッド

        サブクラスでオブジェクトの持つ情報から設定情報に変換する処理を実装する

        Raises:
            NotImplementedError: サブクラスで実装されていない場合

        Returns:
            ReaderArgs: 読み込みプロセスに渡す設定情報
        """
        raise NotImplementedError

    def to_writer_args(self) -> WriterArgs:
        """書き込みプロセスに渡す設定情報に変換するメソッド

        サブクラスでオブジェクトの持つ情報から設定情報に変換する処理を実装する

        Raises:
            NotImplementedError: サブクラスで実装されていない場合

        Returns:
            ReaderArgs: 書き込みプロセスに渡す設定情報
        """
        raise NotImplementedError

    @staticmethod
    def load_product_id(port: str, baud: int) -> str:
        """指定されたポートからproduct_idを読み込むメソッド

        共通であることを確認: A352

        Args:
            port (str): 通信対象のポート

        Raises:
            ConnectionRefusedError: 通信ボーレート全てで通信か失敗した場合

        Returns:
            str: 読み込んだproduct_id
        """
        logger = logging.getLogger(f"{__name__}.load_product")
        command_list: CommandList = [
            [0, 0xFF, 0xFF, 0x0D],  # 復帰おまじない3回
            [0, 0xFF, 0xFF, 0x0D],
            [0, 0xFF, 0xFF, 0x0D],
            [0, 0xFE, 0x01, 0x0D],  # Window 1
            [4, 0x6A, 0x00, 0x0D],  # Product ID 1
            [4, 0x6C, 0x00, 0x0D],  # Product ID 2
            [4, 0x6E, 0x00, 0x0D],  # Product ID 3
            [4, 0x70, 0x00, 0x0D],  # Product ID 4
        ]

        # コマンド送信
        try:
            result = send(command_list, port, baud, timeout=1)

            # 製品型番文字列合成
            #   返却は [Addr, Prod2, Prod1, CR] の構成
            product_id = ""
            for i in range(0, len(result), 4):
                product_id += chr(result[i + 2])  # Prod1 を追加
                product_id += chr(result[i + 1])  # Prod2 を追加
            logger.debug(f"load product id: {product_id}")
            return product_id

        except TimeoutError:
            logger.debug(f"Failed: port={port} baud={baud}")

        # 通信失敗
        raise ConnectionRefusedError(f"Failed to connect: {port}")


class SensorDeviceError(Exception):
    pass
