# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import multiprocessing
import signal
from typing import Any, Callable, Iterable, Mapping, Optional

from logger.log import LoggingClient, LoggingService


def prepare_process() -> None:
    """ロガーが動く環境のプロセス起動方式をspawnに変更する関数

    この関数を実行関数の先頭で呼び出すこと
    """
    multiprocessing.set_start_method("spawn")


class ApplicationProcess(multiprocessing.Process):
    """アプリケーション特有の処理を割り込ませたmultiprocessing.Processのラッパークラス

    以下の機能を持つ
    - プロセスセーフなファイル出力ロギング
    - キーボードインタラプトでプロセスを終了させない
    - Terminateでプロセスを終了させない
      - process.terminate()でプロセスが終了しなくなるため、必ず終わることが保証された関数を渡すこと

    Args:
        multiprocessing.Processに同じ
    """

    def __init__(
        self,
        group: None = None,
        target: Optional[Callable] = None,
        name: Optional[str] = None,
        args: Iterable[Any] = (),
        kwargs: Mapping[str, Any] = {},
        *,
        daemon: Optional[bool] = None,
    ) -> None:
        logging_service = LoggingService.get_instance()
        client = logging_service.get_logging_client()
        super().__init__(
            target=ApplicationProcess._run_target,
            args=(client, target, args, kwargs),
            group=group,
            name=name,
            kwargs=kwargs,
            daemon=daemon,
        )

    @staticmethod
    def _run_target(
        client: LoggingClient,
        target: Optional[Callable] = None,
        args: Iterable[Any] = (),
        kwargs: Mapping[str, Any] = {},
    ):
        # このプログラムではサブプロセスはSIG TERMで終了させない
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        # Ctrl^Cでの終了シグナルが子プロセスに伝搬するため無視する
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        client.configure(client.queue)
        if target:
            target(*args, **kwargs)
