# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from typing import List


def to_int8(b1: int) -> int:
    """
    1 byte -> 8bit signed int 変換
    """
    return (0x80 & b1) * -1 + (0x7F & b1)


def to_uint16(b1: int, b2: int) -> int:
    """
    2 byte -> 16bit unsigned int 変換
    """
    return b1 * 2**8 + b2


def to_int16(b1: int, b2: int) -> int:
    """
    2 byte -> 16bit signed int 変換
    """
    num = to_uint16(b1, b2)
    return (0x8000 & num) * -1 + (0x7FFF & num)


def to_uint32(ba4: List[int]) -> int:
    """
    4 byte -> 32bit unsigned int 変換
    """
    return ba4[0] * 2**24 + ba4[1] * 2**16 + ba4[2] * 2**8 + ba4[3]


def to_int32(ba4: List[int]) -> int:
    """
    4 byte -> 32bit signed int 変換
    """
    num = to_uint32(ba4)
    return (0x80000000 & num) * -1 + (0x7FFFFFFF & num)


def to_dec24(b1: int, b2: int, b3: int) -> float:
    """
    3 bytes -> 24bit signed decimal 変換
    仕様：
      - 24bit, 2の補数
        - bit23 : 符号部 -- 整数部も兼ねると解釈
        - bit22 : 整数部
        - bit21-0 : 小数部
    """
    # 最上位2bit取り出し
    msb2 = (b1 & 0x00C0) >> 6
    # 小数部合成
    dec = ((b1 & 0x3F) << 16) + (b2 << 8) + b3
    # 合成
    return (msb2 & 0b10) * -1 + (msb2 & 0b01) + dec / 2**22


def to_dec32(ba4: List[int]) -> float:
    """
    4 bytes -> 32bit signed decimal 変換
    仕様：
      - 32bit, 2の補数
        - bit31 : 符号部 -- 整数部も兼ねると解釈
        - bit30-24 : 整数部
        - bit23-0  : 小数部
    """
    return to_int32(ba4) / 2**24
